/*****************************************************************************/
/*
                                 mondesi.js

Relies heavily on functionality provided by WZ_JSGRAPHICS.JS (thanks Walter).


COPYRIGHT
---------
Copyright (C) 2008-2013 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION
-------
24-NOV-2013  MGD  migrate functionality from executable to JavaScript
                  optional 15 minute graph width
31-MAR-2013  MGD  per-process monitoring
22-JAN-2013  MGD  finally discovered what garbles printable WZ graphics on
                    MSIE (or at least how to prevent it) - reset "jg_fast"
06-JAN-2013  MGD  refinement (perhaps rework is closer)
13-OCT-2010  MGD  refinement
08-JUL-2010  MGD  add WebSocket support
29-NOV-2008  MGD  initial
*/
/*****************************************************************************/

// versions of MONDESI.EXE this JavaScript is compatible with
var mdsCompatibleVersions = new Array ('5.0.0','5.0.1'); 

////////////////////////
// configurable items //
////////////////////////

var mdsLangClear = 'Clear',
    mdsLangCluster = 'cluster',
    mdsLangConfigure = 'Configure',
    mdsLangConfigAcross = 'node(s) across window',
    mdsLangConfigIgnore = 'ignore WebSocket',
    mdsLangConfigView = 'minutes',
    mdsLangDisk = 'disk',
    mdsLangDrag = 'Drag link to Bookmarks.',
    mdsLangFavs = 'Right-click to add to Favorites.',
    mdsLangView = '5\u2713 15',
    mdsLangView2 = '15\u2713 5',
    mdsLangLink = 'Link',
    mdsLangHistory = 'history',
    mdsLangMonitor = 'Monitor',
    mdsLangPause = 'pause',
    mdsLangPause2 = 'resume',
    mdsLangPrint = 'print',
    mdsLangProc = 'proc',
    mdsLangReset = 'Reset',
    mdsLangScale = 'scale',
    mdsLangStalled = 'STALLED',
    mdsLangStopped = 'STOPPED',
    mdsLangZero = 'zero';

var mdsLangIncompatible = 'JavaScript and executable are incompatible!\n' +
                          'Reload and/or clear browser cache.';

var mdsLangBookmarklet = 'Bookmarklet error!';

var mdsBarCol1 = 'LimeGreen',
    mdsBarCol2 = 'Chocolate',
    mdsBarColAve = 'Yellow',
    mdsHistCol1 = 'DarkGreen',
    mdsHistCol2 = 'Sienna',
    mdsHistColAve = 'GoldenRod',
    mdsObviousBtn = 'Gold';

var mdsOnError = false;

////////////////////////////
// general infrastructure //
////////////////////////////

var mdsMSIEversion = 0;

if (navigator.appName == 'Microsoft Internet Explorer')
{
   var ua = navigator.userAgent;
   var re  = new RegExp('MSIE ([0-9]{1,}[\.0-9]{0,})');
   // MSIE sacrifice speed for correct rendering
   if (re.exec(ua) != null) mdsMSIEversion = parseFloat(RegExp.$1);
}

// only available in later MSIEs (10+)
if (!('indexOf' in Array.prototype)) {
    Array.prototype.indexOf= function(find, i /*opt*/) {
        if (i===undefined) i= 0;
        if (i<0) i+= this.length;
        if (i<0) i= 0;
        for (var n= this.length; i<n; i++)
            if (i in this && this[i]===find)
                return i;
        return -1;
    };
}

var mdsStandAlone = window.location.search.match(/\monitor=2/g) != null;

// dynamically modify a stylesheet
// -------------------------------

function mdsCSS (selector, property, value)
{
    for (var idx = 0; idx < document.styleSheets.length; idx++)
    {
       var sheet = document.styleSheets[idx];
       try {
          var rule = selector + '{' + property + ':' + value + ';}';
          sheet.insertRule(rule,sheet.cssRules.length);
       }
       catch(err) {
          try {
             sheet.addRule(selector,property+':'+value);
          }
          catch(err) {}
       }
    }
}

// change the window title
// -----------------------

mdsTitleArray = new Array();
mdsTitleTitle = null;

function mdsAddNodeToTitle (node)
{
   if (!mdsTitleTitle) mdsTitleTitle = document.title;
   mdsTitleArray.push(node);
   mdsTitleArray.sort(function(s1,s2){return s1.localeCompare(s2)});
   document.title = '';
   for (var idx = 0; idx < mdsTitleArray.length; idx++)
      document.title += mdsTitleArray[idx] + '\u00a0~\u00a0';
   document.title += mdsTitleTitle;
}

// make a bookmarklet
// ------------------

function mdsConfigLink (change)
{
   var cl = document.getElementById('config_link');
   if (!change)
   {
      if (cl.style.display == 'inline-block')
      {
         cl.style.display = 'none';
         return false;
      }
      cl.style.display = 'inline-block';
   }
   if (cl.style.display != 'inline-block') return false;

   var params = 'status=0,toolbar=0,location=0,menubar=0,resizable=0,' +
                'scrollbars=0,height=100,width=100';
   var href = 'javascript:void(window.open("' + mdsConfigUrl() +
                               '","_blank","' + params + '"))';
   var title = 'MONDESI ' + mdsExeVersion;

   var nodes = document.getElementById('nodes').value;
   if (nodes.length < 10)
      cl.style.display = 'none';
   else
   {
      ca = document.getElementById('config_anchor');
      ca.href = ca.innerHTML = href;
      cl.style.display = 'inline-block';
   }

   return false;
}

// generate a URL from the configuration form
// ------------------------------------------

function mdsConfigUrl ()
{
   var action = document.getElementById('config_form').action;
   var nodes = document.getElementById('nodes').value;
   var across = document.getElementById('across');
   across = across[across.selectedIndex].text;
   var view = document.getElementById('view');
   view = view[view.selectedIndex].value;
   var nows = document.getElementById('nows');
   if (nows != undefined && nows.checked) nows = nows.value; else nows = null;
   nodes = nodes.replace(/\n|\r\n|\r/g,'+');
   nodes = nodes.replace(/^\+*/,'').replace(/^\s*/,'')
   nodes = nodes.replace(/\+*$/,'').replace(/\s*$/,'')
   var url = action + '?nodes=' + nodes;
   if (across != 1) url += '&across=' + across;
   if (view) url += '&view=' + view;
   if (nows != undefined) url += '&nows=' + nows;
   url += '&monitor=2';
   return url;
}

function mdsConfigReset ()
{
   document.getElementById('config_form').reset();
   return mdsConfigLink(true);
}

function mdsConfigClear ()
{
   document.getElementById('config_form').reset();
   document.getElementById('nodes').innerHTML = '';
   return mdsConfigLink(true);
}

///////////////
// main page //
///////////////

// main (iframe-containing) page
// -----------------------------

function mdsMainPage ()
{
   if (mdsCompatibleVersions.indexOf(mdsExeVersion) == -1)
   {
      alert(mdsLangIncompatible);
      throw new Error(mdsLangIncompatible);
   }

   if (mdsStandAlone)
   {
      var cf = document.getElementById('form_config');
      if (cf) cf.style.display = 'none';
   }

   if (mdsStandAlone) mdsOpenMainPage();

   document.getElementById('page_configure').value = mdsLangConfigure;
   try {
      var mdsMessageListener =
          window.addEventListener('message',mdsMsgChannel,false);
   } catch(err) { }
}

// when open()ed MonDeSi
// ---------------------

function mdsOpenMainPage ()
{
   window.onresize = function () { mdsFitWindow() };
   if (mdsOnError)
      window.onerror = function (msg,url,lnum)
                                { alert (msg+' '+url+' '+lnum) };

   var mb = document.getElementById('mondesi_body');
   mb.setAttribute('class','mondesi_body2');
   var  nd;
   for (var nnum = 1; nnum; nnum++)
   {
      if (!(nd = document.getElementById('node'+nnum))) break;
      nd.setAttribute('class','node_iframe2');
      if (mdsMSIEversion) nd.setAttribute('src','javascript:void(0);');
   }
   window.resizeBy(1,1);
}

// message posting browsers (HTML5)
// -------------------------------

function mdsMsgChannel (msg)
{
   if (msg.data.substring &&
       msg.data.substring(0,3) == "mds")
   {
      eval(msg.data);
      return;
   }
}

// inter-frame kludge
// ------------------

function mdsBackChannel (forEval)
{
   if (forEval.substring(0,4) == "#mds")
   {
      eval(forEval.substr(1));
      return;
   }
   if (forEval != "") alert("ERROR:"+forEval.substr(1));
}

// remotely called via IFC
// -----------------------

function mdsIframeHeight (iframeId,toTop,width,height)
{
   var ifid = document.getElementById(iframeId);
   ifid.width = ifid.style.width = width + "px";
   ifid.height = ifid.style.height = height + "px";
   if (toTop)
   {
      try {
         if (ifid.offsetTop > 100)
             window.scrollTo(0,ifid.offsetTop-10);
      } catch (err) { }
   }

   if (mdsStandAlone) mdsFitWindow();
}

// open an independent MonDeSi window
// ----------------------------------

function openMonDeSi (url)
{
   if (url)
      window.onerror = function (msg,url,lnum) { alert (msg+' '+url+' '+lnum) };
   else
   {
      if (!document.getElementById('config_form'))
      {
         alert(mdsLangBookmarklet);
         return;
      }
   }

   var name = new Date();
   name = 'mondesi_' + name.getTime();

   var specs = 'toolbar=0,location=0,directories=0,status=0,' +
               'menubar=0,scrollbars=1resizable=0,copyhistory=0,' +
               'width=100,height=100';

   window.open((url?url:mdsConfigUrl()),name,specs);

   return false;
}

// fit the open()ed window to the iframes(s)
// -----------------------------------------
// The window size changes with elements' being made hidden and visible.
// This takes some time for the browser to process and so this function
// provides a succession of "fits" over a period of some 600mS to get the
// window to its final size.  Redundant resizes are avoided.

var mdsFitHeight = 0,
    mdsFitMsec = 0,
    mdsFitOpening = 3,
    mdsFitQuantum = 80,
    mdsFitTimer = null,
    mdsFitWidth = 0;

function mdsFitWindow (timed)
{
   var  cnt, nd,
        height = 0,
        width = 0;

   if (timed)
   {
      mdsFitTimer = null;
      if (mdsFitMsec > 600)
      {
         mdsFitMsec = 0;
         return;
      }
      mdsFitTimer = setTimeout('mdsFitWindow(true)',
                               (mdsFitMsec += mdsFitQuantum));
   }
   else
   if (mdsFitMsec)
   {
      /* a non-timed fit so just restart the sequence */
      if (mdsFitTimer) clearTimeout(mdsFitTimer);
      mdsFitTimer = setTimeout('mdsFitWindow(true)',
                               (mdsFitMsec = mdsFitQuantum));
      return;
   }
   else
      mdsFitTimer = setTimeout('mdsFitWindow(true)',
                               (mdsFitMsec = mdsFitQuantum));

   if (mdsFitOpening)
   {
      // defer the opening resize just a tad (for at least Chrome)
      mdsFitOpening--;
      return;
   }

   for (cnt = 1; cnt; cnt++)
   {
      if (!(nd = document.getElementById('node'+cnt))) break;
      width = Math.max(nd.scrollWidth,nd.offsetWidth,nd.clientWidth);
      height += Math.max(nd.scrollHeight,nd.offsetHeight,nd.clientHeight);
   }
   width += 10;
   height += 10;

   // plus window accoutrements
   width += (window.outerWidth - window.innerWidth);
   height += (window.outerHeight - window.innerHeight);

   if (mdsFitWidth == width && mdsFitHeight == height) return;
   mdsFitWidth = width;
   mdsFitHeight = height;

   var ht = document.getElementsByTagName('html')[0];
   if (mdsFitHeight >= screen.availHeight)
      ht.style.overflowY = 'scroll';
   else
      ht.style.overflowY = 'hidden';

   window.resizeTo(mdsFitWidth,mdsFitHeight);
}

//////////////////////////
// iframe functionality //
//////////////////////////

// note: some storage items are instantiated by MONDESI.C

var mdsData = new Array();

var mdsAlertSecs = 90,  // ninety seconds
    mdsBarHeight = 14,
    mdsBarWidth = 610,
    mdsBody = null,
    mdsClusterVisible = false,
    mdsDiskMount = 0,
    mdsDiskVisible = false,
    mdsGraphHeight = 100,
    mdsGraphWidth = 300,
    mdsHistoryWidth = 0,
    mdsInitialised = false,
    mdsNodeHistVisible = false,
    mdsPauseInProgress = false,
    mdsPauseStopTimer = null,
    mdsPauseStopSecs = 300,   // five minutes
    mdsPerProcPid = 0,
    mdsPerProcVisible = false,
    mdsPerProcHistVisible = false,
    mdsPrevTimeStamp = 0,
    mdsPrintInProgress = false,
    mdsPrintTimeout = 30,  // seconds
    mdsRunning = false,
    mdsScaleTimer = null,
    mdsScaleSecs = 300,  // five minutes
    mdsTimeStamp = 0,
    mdsTopProcVisible = false,
    mdsUpdateDisplay = false,
    mdsUpdatePause = false,
    mdsWatchDogBarks = 30,  // seconds
    mdsWatchDogSecs = 10,
    mdsWatchDogStamp = 0,
    mdsWzPrintable = false;

var mdsNodeMetrics = [ 'cpu','mem','pgfl','flts','bio',
                       'dio','lock','ni','disk','prc' ]; 

var mdsScsMetrics = [ 'scsmg','scskb','mscp' ];

var mdsProcMetrics = [ 'cpu','wset','pgfl','pgs','bio','dio','flts' ]; 

var mdsAlertSpan = '<span style="font-size:140%;font-weight:bold;' +
                   'margin:5px;float:right;color:red;">';
var mdsAlertEndSpan = '</span>';

var mdsAlertText = mdsAlertSpan + mdsLangStalled + mdsAlertEndSpan;

var mdsPauseStopText = mdsAlertSpan + mdsLangStopped + mdsAlertEndSpan;

var mdsProcTable = '<table id="top_proc_table" class="top_proc_table">' +
                   '<tr><th>CPU</th><th>S+E+K</th><th>FLTS</th><th>BIO</th>' +
                   '<th>DIO</th></tr>';
var mdsProcEndTable = '</table>';

var mdsPerProcTable = '<table id="per_proc_table" class="per_proc_table" ' +
                      'onclick="mdsPerProcClick(0)">' +
                      '<tr><td></td><td></td><td></td><td></td><td></td></tr>';
var mdsPerProcEndTable = '</table>';

// allow for the border width in the history graphs
mdsBarWidth += 2;

// start monitoring a node
// -----------------------

function mdsMonitorNode ()
{
   if (mdsCompatibleVersions.indexOf(mdsExeVersion) == -1)
      throw new Error(mdsLangIncompatible);

   if (!mdsInitialised)
   {
      mdsBody = document.getElementsByTagName('body').item(0);

      mdsHistoryWidth = mdsGraphWidth;

      mdsBuildNodeGraphs();
      mdsBuildControlBtns();
      mdsSetDisplayHeight ();

      setInterval('mdsWatchDog()',mdsWatchDogSecs*1000);

      if (mdsConfigView == 15) document.getElementById('btn_view').click();

      if (mdsStandAlone)
         if (mdsOnError)
            window.onerror = function (msg,url,lnum)
                                      { alert (msg+' '+url+' '+lnum) };

      // delay until after the data stream is established
      var msg = 'mdsToMsgChannel(' + "'" + 'mdsAddNodeToTitle("' +
                mdsNodeName + '")' + "'" + ')';
      setTimeout(msg,2000);

      mdsInitialised = true;
   }

   mdsStreamData (true);
   mdsScaleTimer = setTimeout('mdsScale()',mdsScaleSecs*1000);
   mdsRunning = mdsUpdateDisplay = true;
}

// watch that we're not disconnected
// ---------------------------------

function mdsWatchDog ()
{
   if (mdsTimeStamp > mdsPrevTimeStamp)
      mdsWatchDogStamp = mdsTimeStamp;
   else
      mdsWatchDogStamp += mdsWatchDogSecs;
   mdsPrevTimeStamp = mdsTimeStamp;
   // if no time-stamp update in this period attempt to reconnect
   if (mdsWatchDogStamp >= mdsTimeStamp + mdsWatchDogBarks)
      mdsStreamData (true);
}

// the MonDeSi executable has (re)started
// --------------------------------------

function mdsExecStartup ()
{
   mdsProcDisplay();
   mdsPerProcDisplay();
}

////////////////////////
// build data display //
////////////////////////

// build basic bar and history graphs
// ----------------------------------

function mdsBuildNodeGraphs ()
{
   var nb = document.getElementById('node_bars');
   var nh = document.getElementById('node_hist');
   for (var cnt = 0; mdsNodeMetrics[cnt]; cnt++)
   {
      var which = 'node_' + mdsNodeMetrics[cnt];
      var bar = nb.appendChild(mdsBuildBar(which));
      nh.appendChild(mdsBuildHist(which));
      mdsDataInit(which,true);
   }
   if (mdsCluster)
   {
      for (var cnt = 0; mdsScsMetrics[cnt]; cnt++)
      {
         var which = 'node_' + mdsScsMetrics[cnt];
         var bar = nb.appendChild(mdsBuildBar(which));
         nh.appendChild(mdsBuildHist(which));
         mdsDataInit(which,true);
      }
   }
   for (var cnt = 0; cnt < mdsDiskCount; cnt++)
   {
      var which = 'node_disk_' + cnt;
      var bar = nb.appendChild(mdsBuildBar(which));
      nh.appendChild(mdsBuildHist(which));
      mdsDataInit(which,true);
   }
   // set the bottom border of the last visible bar
   var cn = document.getElementById('node_bars').childNodes;
   for (var cnt = cn.length-1; cnt > 0; cnt--)
   {
      if (cn[cnt].style.display == 'none') continue;
      cn[cnt].style.borderBottomWidth = '1px';
      break;
   }
   var pph = document.getElementById('per_proc_hist');
   for (var cnt = 0; mdsProcMetrics[cnt]; cnt++)
   {
      var which = 'proc_' + mdsProcMetrics[cnt];
      pph.appendChild(mdsBuildHist(which,'per_proc'));
      mdsDataInit(which,true);
   }
   var div = document.createElement('div');
   div.setAttribute('id','inter_hist_gap');
   pph.appendChild(div);

   var hnm = document.getElementById('hist_node_mem');
   var hnp = document.getElementById('hist_node_pgfl');
   hnm.setAttribute('onmouseover',
                    'mdsForeFront("hist_node_pgfl","hist_node_mem",true)');
   hnp.setAttribute('onmouseout',
                    'mdsForeFront("hist_node_mem","hist_node_pgfl")');

   var ppw = document.getElementById('hist_proc_wset');
   var ppp = document.getElementById('hist_proc_pgfl');
   ppw.setAttribute('onmouseover',
                    'mdsForeFront("hist_proc_pgfl","hist_proc_wset")');
   ppp.setAttribute('onmouseout',
                    'mdsForeFront("hist_proc_wset","hist_proc_pgfl",true)');
}

function mdsBuildBar (id)
{
   var ced = document.createElement('div');
   ced.setAttribute('id','bar_'+id);
   ced.setAttribute('class','node_bar');
   if (id == 'node_pgfl' ||
       id.substr(0,8) == 'node_scs' ||
       id.substr(0,9) == 'node_mscp' ||
       id.substr(0,10) == 'node_disk_')
      ced.style.display = 'none';
   return ced;
}

function mdsBuildHist (id)
{
   var ced = document.createElement('div');
   ced.setAttribute('id','hist_'+id);
   ced.setAttribute('class','node_hist');
   if (id == 'proc_pgfl' ||
       id == 'node_pgfl' ||
       id.substr(0,8) == 'node_scs' ||
       id.substr(0,9) == 'node_mscp' ||
       id.substr(0,10) == 'node_disk_')
      ced.style.display = 'none';
   return ced;
}

function mdsBuildControlBtns ()
{
   var btn;
   var nc = document.getElementById('node_control');
   nc.appendChild(mdsBuildOneButton(mdsLangHistory,'mdsHistoryClick()'));
   nc.appendChild(btn = mdsBuildOneButton(mdsLangCluster,'mdsClusterClick()'));
   if (!mdsCluster) btn.disabled = true;
   nc.appendChild(mdsBuildOneButton(mdsLangDisk,'mdsDiskClick()'));
   nc.appendChild(btn = mdsBuildOneButton(mdsLangProc,'mdsProcClick()'));
   if (!mdsTopProc) btn.disabled = true;
   nc.appendChild(btn = mdsBuildOneButton('pause','mdsPauseClick()'));
   btn.value = mdsLangPause;

   var br = document.createElement('br');
   nc.appendChild(br);

   nc.appendChild(mdsBuildOneButton(mdsLangScale,'mdsScale()'));
   nc.appendChild(mdsBuildOneButton(mdsLangZero,'mdsZeroData()'));
   nc.appendChild(mdsBuildOneButton(mdsLangPrint,'mdsPrintClick()'));
   nc.appendChild(btn = mdsBuildOneButton('close','parent.window.close()'));
   if (!mdsStandAlone) btn.disabled = true;
   nc.appendChild(btn = mdsBuildOneButton('view','mdsViewClick()'));
   btn.value = mdsLangView;

   document.getElementById('btn_cluster').disabled = !mdsCluster;
   document.getElementById('btn_disk').disabled = (mdsDiskCount == 0);
}

function mdsBuildOneButton (id,onclick)
{
   var btn = document.createElement('input');
   btn.setAttribute('type','submit');
   btn.setAttribute('id','btn_'+id);
   btn.setAttribute('class','btn');
   btn.setAttribute('value',id);
   btn.setAttribute('onclick',onclick);
   return btn;
}

function mdsConfigOptions ()
{
   var ce = document.getElementById('config_view');
   ce.innerHTML = mdsLangView;
   if (WASDwebsocket && 'WebSocket' in window)
      document.write('<br><input id=\"nows\" type=\"checkbox\" name=\"nows\" ' +
                     'value=\"1\" onchange="mdsConfigLink(true)"' +
                     (mdsConfigNOws ? ' checked' : '') +
                     '>&nbsp;ignore WebSocket</span>');
   document.getElementById('config_across').innerHTML = mdsLangConfigAcross;
   document.getElementById('config_view').innerHTML = mdsLangConfigView;
   setTimeout('mdsConfigOptions2()',100);
}

function mdsConfigOptions2 ()
{
   document.getElementById('config_monitor').value = mdsLangMonitor;
   document.getElementById('config_link').value = mdsLangLink;
   document.getElementById('config_clear').value = mdsLangClear;
   document.getElementById('config_reset').value = mdsLangReset;
   document.getElementById('config_drag').innerHTML =
      mdsMSIEversion ? mdsLangFavs : mdsLangDrag;
}

/////////////////////
// data management //
/////////////////////

// (re)initialise item data
// ------------------------

function mdsDataInit (which,active)
{
   if (mdsData[which] == undefined)
   {
      mdsData[which] = new Object();
      mdsData[which].timer = null;
      mdsData[which].which = which;
      var title = which.substr(which.indexOf('_')+1);
      if (title.substr(0,4) == 'scs_')
         mdsData[which].title = title.substr(4).toUpperCase();
      else
         mdsData[which].title = title.toUpperCase();
      mdsData[which].active = active;
      mdsData[which].line2 = mdsData[which].line3 = '';
      if (which.substr(0,5) == 'node_')
         mdsData[which].bar_jsgr = new jsGraphics('bar_'+which);
      mdsData[which].hist_jsgr = new jsGraphics('hist_'+which);
   }
   else
   {
      delete mdsData[which].hist1;
      delete mdsData[which].hist2;
   }

   mdsData[which].ave = mdsData[which].max =
      mdsData[which].val1 = mdsData[which].val2 =
      mdsData[which].scale = 0;
   mdsData[which].fixed = false;
   mdsData[which].visible = false;
   mdsData[which].hist1 = new Array();
   mdsData[which].hist2 = new Array();

   if (which.substr(0,5) == 'node_')
   {
      mdsBarGraph (which);
      mdsBarPaint (which);
   }

   mdsHistGraph (which);
   mdsHistPaint (which);
}

// set and repaint the title
// -------------------------

function mdsDataTitle (which,title,line2,line3)
{
   mdsData[which].title = title;
   mdsData[which].line2 = line2;
   mdsData[which].line3 = line3;
   mdsDatum(which);
}

// scale displays according to maximum datum
// -----------------------------------------

function mdsScale ()
{
   for (which in mdsData)
   {
      if (mdsData[which].fixed) continue;
      var hist1 = mdsData[which].hist1;

      var max = 0;
      for (tdata in hist1)
         if (hist1[tdata] > max) max = hist1[tdata];
      mdsData[which].max = mdsData[which].scale = max;
      mdsData[which].ave = 0;
      if (which.substr(0,5) == 'node_') mdsBarPlot(which);
      mdsHistPlot(which);
   }

   clearTimeout(mdsScaleTimer);
   mdsScaleTimer = setTimeout('mdsScale()',mdsScaleSecs*1000);
}

// zero data, reset graphs
// -----------------------

function mdsZeroData ()
{
   for (which in mdsData)
   {
      mdsDataInit(which);
      mdsDatum(which,0);
   }
   mdsHistoryWidth = mdsGraphWidth;
}

// record a datum and graph its value
// ----------------------------------
// (called by the data streaming from the MonDeSi executable)

function mdsDatum (which,scale,value1,value2)
{
   if (mdsData[which] == undefined) return;

   if (!mdsData[which].active) return;

   if (scale == undefined) scale = -1;
   if (value1 == undefined) value1 = -1;
   if (value2 == undefined) value2 = -1;

   // arrange a timed update (must be done before 'scale' is adjusted)
   var updateGraph = 'mdsDatum("'+which+'",'+scale+','+value1+','+value2+')';
   clearTimeout(mdsData[which].timer);
   mdsData[which].timer = setTimeout(updateGraph,mdsRefreshInterval*2000);

   if (value1 > mdsData[which].max) mdsData[which].max = value1;
   var max = mdsData[which].max;
   if (scale >= 0)
      mdsData[which].fixed = true;
   else
      scale = max;
   mdsData[which].scale = scale;

   // calculate an average for most data
   var average = 0;
   if (which != 'node_mem' && which != 'node_pgfl' && which != 'node_prc')
   {
      var hist1 = mdsData[which].hist1;
      var count = 0;
      var total = 0;
      for (tdata in hist1)
      {
         count++;
         total += hist1[tdata];
      }
      if (count > 10) average = Math.round((total / count));
      if (average < 0) average = 0;
   }

   mdsData[which].val1 = value1;
   mdsData[which].val2 = value2;
   mdsData[which].ave = average;

   // record the datum in a sparse (js hash) array indexed by timestamp
   var tdata = Math.round(mdsTimeStamp).toString();
   if (value1 >= 0) mdsData[which].hist1[tdata] = value1;
   if (value2 >= 0) mdsData[which].hist2[tdata] = value2;

   if (which.substr(0,5) == 'node_') mdsBarPlot(which);

   mdsHistPlot(which);
}

// set the process table
// ---------------------

function mdsMonProc (states,procs)
{
   if (!mdsUpdateDisplay) return;

   var procTableDiv = document.getElementById('node_process');
   procTableDiv.innerHTML = null;
   procTableDiv.innerHTML = '<div class="proc_states">' + states + '</div>' +
                            mdsProcTable + procs + mdsProcEndTable;

}

// make all PIDs clickable
// -----------------------

function mdsMonProcPid ()
{
   if (!mdsUpdateDisplay) return;

   var table = document.getElementById('top_proc_table');
   for (var i = 0, row; row = table.rows[i]; i++) {
      for (var j = 0, col; col = row.cells[j]; j++) {
         var pid = col.innerHTML.substr(0,8);
         if (pid.length == 8) col.onclick = mdsPidClick;
      }  
   }
}

// update the textual node information
// -----------------------------------

function mdsNodeNotes (text)
{
   if (!mdsUpdateDisplay) return;

   var nodeNotes = document.getElementById('node_notes');
   nodeNotes.innerHTML = null;
   nodeNotes.innerHTML = text;
   // each minute the alert gets recheduled if updating is occuring
   mdsResetAlert(true);
}

// response to a PID click
// -----------------------

function mdsPidClick (event)
{ 
   var pid = event.currentTarget.innerHTML.substr(0,8);
   mdsPerProcClick (pid);
   event.cancelBubble = true;  // MSIE
   if (event.stopPropagation) event.stopPropagation();
}

// prompt for an ad hoc PID
// ------------------------

function mdsPidPrompt (event)
{
   pid = prompt('Enter PID','');
   if (!isNaN(parseInt(pid,16))) mdsPerProcClick(pid);
   event.cancelBubble = true;  // MSIE
   if (event.stopPropagation) event.stopPropagation();
}

// per-process table
// -----------------

function mdsMonPerProc (table)
{
   if (!mdsUpdateDisplay) return;

   var ppp = document.getElementById('per_proc_pid');
   ppp.innerHTML = null;
   ppp.innerHTML = mdsPerProcTable + table + mdsPerProcEndTable;
}

///////////////////
// button clicks //
///////////////////

// cluster button has been toggled
// -------------------------------

function mdsClusterClick (setValue)
{
   mdsClusterVisible = !mdsClusterVisible;
   var cn = document.getElementById('node_bars').childNodes;
   for (var cnt = 0; cnt < cn.length; cnt++)
      if (cn[cnt].id.substr(0,12) == 'bar_node_scs' ||
          cn[cnt].id.substr(0,13) == 'bar_node_mscp')
         if (mdsClusterVisible)
            cn[cnt].style.display = 'block';
         else
            cn[cnt].style.display = 'none';
   for (var cnt = cn.length-1; cnt > 0; cnt--)
      cn[cnt].style.borderBottomWidth = 0;
   for (var cnt = cn.length-1; cnt > 0; cnt--)
   {
      if (cn[cnt].style.display == 'none') continue;
      cn[cnt].style.borderBottomWidth = '1px';
      break;
   }
   var cn = document.getElementById('node_hist').childNodes;
   for (var cnt = 0; cnt < cn.length; cnt++)
      if (cn[cnt].id.substr(0,13) == 'hist_node_scs' ||
          cn[cnt].id.substr(0,14) == 'hist_node_mscp')
         if (mdsClusterVisible)
            cn[cnt].style.display = 'inline-block';
         else
            cn[cnt].style.display = 'none';
   mdsSetDisplayHeight();
   return false;
}

// disk button has been toggled
// ----------------------------

function mdsDiskClick ()
{
   mdsDiskVisible = !mdsDiskVisible;
   var cn = document.getElementById('node_bars').childNodes;
   for (var cnt = 0; cnt < cn.length; cnt++)
      if (cn[cnt].id.substr(0,14) == 'bar_node_disk_')
         if (mdsDiskVisible)
            cn[cnt].style.display = 'block';
         else
            cn[cnt].style.display = 'none';
   for (var cnt = cn.length-1; cnt > 0; cnt--)
      cn[cnt].style.borderBottomWidth = 0;
   for (var cnt = cn.length-1; cnt > 0; cnt--)
   {
      if (cn[cnt].style.display == 'none') continue;
      cn[cnt].style.borderBottomWidth = '1px';
      break;
   }
   var cn = document.getElementById('node_hist').childNodes;
   for (var cnt = 0; cnt < cn.length; cnt++)
      if (cn[cnt].id.substr(0,15) == 'hist_node_disk_')
         if (mdsDiskVisible)
            cn[cnt].style.display = 'inline-block';
         else
            cn[cnt].style.display = 'none';
   mdsSetDisplayHeight();
   return false;
}

// control use of the [proc] button
// --------------------------------

function mdsProcButton (top,per)
{
///   document.getElementById('btn_proc').disabled = !top;
   if (!top) {
      mdsTopProcVisible = per = false;
      mdsProcDisplay();
      controlTopProc();
   }
   if (!per) {
      mdsPerProcVisible = false;
      mdsPerProcPid = 0;
      mdsPerProcDisplay();
      controlPerProc ();
   }
}

// top-process display
// -------------------

function mdsProcClick (setValue)
{
   if (setValue == undefined) {
      if (!document.getElementById('btn_proc').disabled)
         mdsTopProcVisible = !mdsTopProcVisible;
   }
   else
      mdsTopProcVisible = setValue;
   mdsProcDisplay();
   controlTopProc();
   return true;
}

// set display of top-processes
// ----------------------------

function mdsProcDisplay ()
{
   var tp = document.getElementById('node_process');
   if (mdsTopProcVisible) {
      tp.style.display = 'table-cell';
      tp.innerHTML = null;
      tp.innerHTML = '<div class="proc_states">&nbsp;</div>' +
                      mdsProcTable + mdsProcEndTable;
   }
   else
   if (tp) {
      tp.style.display = 'none';
      tp.innerHTML = null;
      mdsPerProcVisible = mdsPerProcHistVisible = mdsPerProcPid = 0;
   }
   mdsPerProcDisplay ();
   mdsSetDisplayHeight();
}

// per-process display
// -------------------

function mdsPerProcClick (pid)
{
   mdsPerProcVisible = ((mdsPerProcPid = pid) != 0);
   if (!mdsPerProcVisible) mdsPerProcHistVisible = false;

   // zero the per-process data
   for (which in mdsData)
   {
      if (which.substr(0,5) == 'node_') continue;
      mdsDataInit(which);
      mdsDatum(which,0);
   }

   mdsPerProcDisplay();
   mdsPerProcHistDisplay();
   controlPerProc ();
   return true;
}

// set display of per-process
// --------------------------

function mdsPerProcDisplay ()
{
   var ppp = document.getElementById('per_proc_pid');
   if (mdsPerProcVisible) {
      ppp.style.display = 'table-cell';
      ppp.innerHTML = mdsPerProcTable + mdsPerProcEndTable;
   }
   else
   if (ppp) {
      ppp.style.display = 'none';
      ppp.innerHTML = null;
      perProcHistVisible = false;
   }

   mdsPerProcHistDisplay();
   mdsSetDisplayHeight();
}

// per-process history
// -------------------

function mdsPerProcHistClick (event)
{
   if (mdsPerProcHistVisible)
      mdsPerProcHistVisible = false;
   else
      mdsPerProcHistVisible = mdsPerProcVisible;
   mdsPerProcHistDisplay();
   event.cancelBubble = true;  // MSIE
   if (event.stopPropagation) event.stopPropagation();
   return true;
}

// set display of per-process history
// ----------------------------------

function mdsPerProcHistDisplay ()
{
   var pph = document.getElementById('per_proc_hist');
   if (mdsPerProcHistVisible)
      pph.style.display = 'table-cell';
   else
   if (pph)
      pph.style.display = 'none';
   mdsInterHistGap();
   mdsSetDisplayHeight();
}

// history button has been toggled
// -------------------------------

function mdsHistoryClick ()
{
   var nh = document.getElementById('node_hist');
   if (mdsNodeHistVisible = !mdsNodeHistVisible)
      nh.style.display = 'table-cell';
   else
      nh.style.display = 'none';
   mdsInterHistGap();
   mdsSetDisplayHeight();
   for (var which in mdsData) mdsHistPlot(which);
   return false;
}

// adjust inter-history gap
// ------------------------

function mdsInterHistGap ()
{
   var ppg = document.getElementById('inter_hist_gap');
   if (mdsNodeHistVisible && mdsPerProcHistVisible)
      ppg.style.display = 'block';
   else
      ppg.style.display = 'none';
}

// set the display height
// ----------------------

function mdsSetDisplayHeight ()
{
   var nb = document.getElementById('node_block');
   var displayWidth = nb.offsetWidth+5;
   var displayHeight = nb.offsetHeight+15;
   var remotePC = 'mdsIframeHeight("' + mdsIframeId + '",' +
                  mdsNodeHistVisible + ',' +
                  displayWidth + ',' + displayHeight + ')';
   mdsToMsgChannel(remotePC);
}

// pause and release the display update
// ------------------------------------

var mdsPauseBgCol;

function mdsPauseClick ()
{ 
   var btn = document.getElementById('btn_pause');
   if (mdsRunning && !mdsPauseInProgress)
   {
      // pause the data updates
      mdsPauseBgCol = btn.style.backgroundColor;
      btn.style.backgroundColor = mdsObviousBtn;
      btn.value = mdsLangPause2;
      mdsResetAlert(false);
      clearTimeout(mdsScaleTimer);
      mdsScaleTimer = null;
      mdsPauseStopTimer = setTimeout('mdsPauseStop()',mdsPauseStopSecs*1000);
      // just in case the browser print dialogue is engaged whilst stopped
      mdsMakePrintable(true);
      mdsPauseInProgress = true;
      mdsUpdateDisplay = false;
   }
   else
   if (mdsRunning)
   {
      // recommence the data updates
      btn.style.backgroundColor = mdsPauseBgCol;
      btn.value = mdsLangPause;
      mdsResetAlert(true);
      mdsMakePrintable(false);
      mdsProcClick(mdsTopProcVisible);
      clearTimeout(mdsPauseStopTimer);
      mdsPauseStopTimer = null;
      mdsScaleTimer = setTimeout('mdsScale()',mdsScaleSecs*1000);
      mdsPauseInProgress = false;
      if (!mdsPrintInProgress) mdsUpdateDisplay = true;
   }
   else
   {
      // restart data collection and display
      btn.style.backgroundColor = mdsPauseBgCol;
      btn.value = 'pause';
      mdsStreamData (true);
      mdsZeroData();
      if (mdsPrintInProgress) 
         mdsPrintClick();
      else
         mdsMakePrintable(false);
      mdsProcClick(mdsTopProcVisible);
      mdsScaleTimer = setTimeout('mdsScale()',mdsScaleSecs*1000);
      mdsPauseInProgress = false;
      mdsRunning = mdsUpdateDisplay = true;
   }
   return true;
}

// timer-driven stop
// -----------------

function mdsPauseStop ()
{
   // been paused for long enough!
   mdsRunning = mdsUpdateDisplay = false;
   mdsPauseStopTimer = null;
   clearTimeout(mdsScaleTimer);
   mdsScaleTimer = null;
   // shut down the client/server IPC
   mdsStreamData (false);
   mdsPauseInProgress = false;
   var btn = document.getElementById('btn_pause');
   btn.value = 'START';
   // append an informational to the notes
   var nodeNotes = document.getElementById('node_notes');
   var notesText = nodeNotes.innerHTML;
   nodeNotes.innerHTML = null;
   nodeNotes.innerHTML = mdsPauseStopText + notesText;
}

// 5 or 15 minute history graphs
// -----------------------------

function mdsViewClick ()
{ 
   var be = document.getElementById('btn_view');
   if (mdsGraphWidth < 500)
   {
      // once set wide history width is never reeset narrow
      mdsGraphWidth = mdsHistoryWidth = 924;
      be.value = mdsLangView2;
      mdsCSS ('.node_hist', 'display', 'block');
   }
   else
   {
      mdsGraphWidth = 300;
      be.value = mdsLangView;
      mdsCSS ('.node_hist', 'display', 'inline-block');
   }
   mdsCSS ('.node_hist', 'width', mdsGraphWidth+'px');
   mdsSetDisplayHeight();
   for (var which in mdsData) mdsHistPlot(which);
}

// make the first visible the second not
// -------------------------------------

var mdsForeFrontTimeout = new Array();
var mdsForeFrontSeconds = 30;

function mdsForeFront (which1,which2,revert)
{
   document.getElementById(which1).style.display = 'inline-block';
   document.getElementById(which2).style.display = 'none';
   if (mdsForeFrontTimeout[which1]) {
      clearTimeout (mdsForeFrontTimeout[which1]);
      mdsForeFrontTimeout[which1] = null;
   }
   if (revert)
      mdsForeFrontTimeout[which1] =
         setTimeout('mdsForeFront("'+which2+'","'+which1+'")',
                    mdsForeFrontSeconds*1000);
}

// tell script to start/stop top-process monitoring
// ------------------------------------------------

function controlTopProc () {

   if (!ws) {
      // not using WebSocket so just restart the stream
      if (mdsTopProcVisible) mdsStreamData (true);
      return;
   }

   try {
      if (mdsTopProcVisible)
         ws.send ('topproc=1');
      else
         ws.send ('topproc=0');
   }
   catch (err) {
      // stop/start? WebSocket probably in process of being established
      setTimeout('controlTopProc()',1000);
   }

   return false;
}

// tell script to start/stop per-process monitoring
// ------------------------------------------------

function controlPerProc () {

   if (!ws) {
      // not using WebSocket so just restart the stream
      if (mdsPerProcPid) mdsStreamData (true);
      return;
   }

   try {
      ws.send ('perproc='+mdsPerProcPid);
   }
   catch (err) {
      // stop/start? WebSocket probably in process of being established
      setTimeout('controlPerProc()',1000);
   }

   return false;
}

///////////////////////////////
// printing - a juggling act //
///////////////////////////////

var mdsPrintBgCol;
var mdsPrintClickTimer = null;

function mdsPrintClick ()
{
   if (mdsPrintInProgress)
      mdsAfterPrint();
   else
   {
      mdsBeforePrint();
      setTimeout('window.print()',100);
   }
   return true;
}

function mdsBeforePrint ()
{
   if (mdsPrintInProgress) return true;
   mdsPrintInProgress = true;
   var btn = document.getElementById('btn_print');
   if (btn)
   {
      // bloody-mindded MSIE 10(?)
      mdsPrintBgCol = btn.style.backgroundColor;
      btn.style.backgroundColor = mdsObviousBtn;
   }
   clearTimeout(mdsPrintClickTimer);
   mdsPrintClickTimer = null;
   mdsPrintClickTimer = setTimeout('mdsPrintClick()',mdsPrintTimeout*1000);
   mdsResetAlert(false);
   mdsMakePrintable(true);
   mdsUpdateDisplay = false;
   return true;
};

function mdsAfterPrint()
{
   if (!mdsPrintInProgress) return true;
   mdsPrintInProgress = false;
   var btn = document.getElementById('btn_print');
   if (btn)
   {
      // bloody-mindded MSIE 10(?)
      btn.style.backgroundColor = mdsPrintBgCol;
   }
   clearTimeout(mdsPrintClickTimer);
   mdsPrintClickTimer = null;
   mdsResetAlert(true);
   mdsMakePrintable(false);
   if (!mdsPauseInProgress) mdsUpdateDisplay = true;
   return true;
};

// NOTE: this is in-line code
var mdsMatchMedia;
if (window.matchMedia)
{
    mdsMatchMedia = window.matchMedia('print');
    mdsMatchMedia.addListener(function(mql) {
       if (mql.matches)
          mdsBeforePrint();
       else
          mdsAfterPrint();
    });
}
// a la MSIE
window.onbeforeprint = mdsBeforePrint;
window.onafterprint = mdsAfterPrint;

// (re)set WZ printable flag and redraw
// ------------------------------------

function mdsMakePrintable (MakePrint)
{
   // 'jg_fast' is a datum in WZ_JSGRAPHICS.JS and may not exist for MSIE10+
   if (typeof jg_fast != 'undefined')
   {
      var jgFast = jg_fast;
      if (mdsWzPrintable = MakePrint)
      {
         // reset the WZ mode that garbles printable WZ graphics on MSIE
         if (mdsMSIEversion && jgFast) jg_fast = false;
      }
   }
   for (which in mdsData)
   {
      if (which.substr(0,5) == 'node_') mdsBarPlot(which);
      mdsHistPlot(which);
   }
   if (typeof jg_fast != 'undefined') jg_fast = jgFast;
   mdsWzPrintable = false;
}

//////////////
// graphing //
//////////////

// return a string with the value as unit, kilo or mega
// ----------------------------------------------------

function mdsValueString (value)
{
   if (value > 999999999) {
      value = (value / 1073741824).toFixed(1);
      if (value > 99) value = (value/1).toFixed(0);
      var dvalue = value.toString() + 'G';
   }
   else
   if (value > 999999) {
      value = (value / 1048576).toFixed(1);
      if (value > 99) value = (value/1).toFixed(0);
      var dvalue = value.toString() + 'M';
   }
   else
   if (value > 9999) {
      value = (value / 1024).toFixed(1);
      if (value > 99) value = (value/1).toFixed(0);
      var dvalue = value.toString() + 'k';
   }
   else
      var dvalue = value.toString();
   return dvalue;
}

// display a numeric value to the right of a bar or graph
// ------------------------------------------------------

function mdsDrawValue (jsgr,xpos,ypos,color,value)
{
   var dval = mdsValueString(value);
   xpos = xpos - 5 - (dval.length * 8) + (dval.length * 2);
   if (dval.indexOf('k') >= 0) xpos += 2;
   if (dval.indexOf('.') >= 0) xpos += 2;
   jsgr.setColor(color);
   jsgr.drawString(dval,xpos,ypos);
}

// initialise a single datum bar graph
// -----------------------------------

function mdsBarGraph (which)
{
   var jsgr = mdsData[which].bar_jsgr;
   jsgr.recreate();
   jsgr.setPrintable(mdsWzPrintable);
   jsgr.setFont ('arial','10px',Font.PLAIN); 
   if (mdsData[which].active)
      jsgr.setColor('white');
   else
      jsgr.setColor('#d3d3d3'); // 'lightGray' - stupid MSIE
   jsgr.fillRect (0, 0, mdsBarWidth, mdsBarHeight);
}

// add the axes and title of a bar graph then paint
// ------------------------------------------------

function mdsBarPaint (which)
{
   var jsgr = mdsData[which].bar_jsgr;
   jsgr.setColor('black');
   if (which == 'node_cpu' && mdsCpuCount > 1)
      jsgr.drawString(mdsData[which].title+'x'+mdsCpuCount,7,2);
   else
      jsgr.drawString(mdsData[which].title,7,2);
   jsgr.paint();
}

// render a single datum of the bar graph
// --------------------------------------

function mdsBarPlot (which)
{
   if (!mdsUpdateDisplay) return;

   if (mdsData[which] == undefined) return;

   mdsBarGraph(which);
   if (!mdsData[which].active)
   {
      mdsBarPaint(which);
      return;
   }

   var height = mdsBarHeight;

   var jsgr = mdsData[which].bar_jsgr;
   jsgr.setColor('white');
   jsgr.fillRect (0, 0, mdsBarWidth, mdsBarHeight);

   mdsDrawValue(jsgr,mdsBarWidth,2,'black',mdsData[which].scale);

   var widthAve = mdsData[which].ave *
                  mdsBarWidth / mdsData[which].scale;
   if (widthAve > mdsBarWidth) widthAve = mdsBarWidth;
   var widthVal1 = mdsData[which].val1 *
                   mdsBarWidth / mdsData[which].scale;
   if (widthVal1 > mdsBarWidth) widthVal1 = mdsBarWidth;

   if (mdsData[which].val2 >= 0)
      var widthVal2 = mdsData[which].val2 *
                      mdsBarWidth / mdsData[which].scale;
   else
      var widthVal2 = 0;

   if (widthAve > 0 && widthAve > widthVal1 && widthAve > widthVal2)
   {
      jsgr.setColor(mdsBarColAve);
      jsgr.fillRect (0, 0, widthAve, height);
      if (widthAve > 60)
         mdsDrawValue(jsgr,widthAve,2,'black',mdsData[which].ave);
   }

   if (widthVal1 > 0 && widthVal1 > widthVal2)
   {
      jsgr.setColor(mdsBarCol1);
      jsgr.fillRect (0, 0, widthVal1, height);
      if (widthVal1 > 60)
         mdsDrawValue(jsgr,widthVal1,2,'black',mdsData[which].val1);
   }

   if (widthVal2 > 0)
   {
      jsgr.setColor(mdsBarCol2);
      jsgr.fillRect (0, 0, widthVal2, height);
      if (widthVal2 > 60)
         mdsDrawValue(jsgr,widthVal2,2,'black',mdsData[which].val2);
   }

   mdsBarPaint(which);
}

// initialise a history line graph
// -------------------------------

function mdsHistGraph (which)
{
   var jsgr = mdsData[which].hist_jsgr;
   jsgr.recreate();
   jsgr.setPrintable(mdsWzPrintable);
   jsgr.setFont ('arial','10px',Font.PLAIN); 
   if (mdsData[which].hist1 != undefined)
      if (mdsData[which].active)
         jsgr.setColor('white');
      else
         jsgr.setColor('#d3d3d3'); // 'lightGray' - stupid MSIE
   jsgr.fillRect (0, 0, mdsGraphWidth, mdsGraphHeight);
}

// provide the axes and title of the graph then display
// ----------------------------------------------------

function mdsHistPaint (which)
{
   var jsgr = mdsData[which].hist_jsgr;
   jsgr.setColor('black');
   var grad = 30;
   for (var x = mdsGraphWidth; x > 0; x -= 30)
      if ((grad += 30) % 60)
         jsgr.drawLine (x, mdsGraphHeight, x, mdsGraphHeight-2);
      else
         jsgr.drawLine (x, mdsGraphHeight, x, mdsGraphHeight-5);
   jsgr.drawLine (-1, 0, -1, mdsGraphHeight);
   jsgr.drawLine (mdsGraphWidth, 0, mdsGraphWidth, mdsGraphHeight);
   var title = mdsData[which].title;
   if ((which == 'node_cpu' || which == 'proc_cpu') && mdsCpuCount > 1)
      title += 'x' + mdsCpuCount;
   if (which.substr(0,5) == 'proc_') title += '<sub>P</sub>';
   if (mdsData[which].line2.length)
      title += '<br>' + mdsData[which].line2;
   if (mdsData[which].line3.length)
      title += '<br><span style="color:red;">' +
               mdsData[which].line3 + '</span>';
   jsgr.drawString(title,5,5);
   jsgr.paint();
}

// display the numeric scale and values on the plot
// ------------------------------------------------

function mdsHistScale (jsgr,which,value1,value2,scale)
{
   if (scale >= 0)
      mdsDrawValue(jsgr,mdsGraphWidth,5,'black',scale);
   if (scale > 0 && value1 >= 0)
      mdsDrawValue(jsgr,mdsGraphWidth,20,mdsHistCol1,value1);
   if (scale > 0 && value1 > 0 && value2 >= 0)
      mdsDrawValue(jsgr,mdsGraphWidth,35,mdsHistCol2,value2);
}

// display an average on the plot
// ------------------------------

function mdsHistAverage (jsgr,value,scale)
{
   if (value <= 0 || value == NaN) return;
   var average = value * mdsGraphHeight / scale;
   if (average <= 0) return;
   if (average > mdsGraphHeight) average = mdsGraphHeight;
   jsgr.setColor(mdsHistColAve);
   jsgr.drawLine (mdsGraphWidth, mdsGraphHeight-average-1,
                  0, mdsGraphHeight-average-1);
   if (average < 12)
      jsgr.drawString(mdsValueString(value),5,mdsGraphHeight-15-average+2);
   else
   if (average < 85)
      jsgr.drawString(mdsValueString(value),5,mdsGraphHeight-average+2);
   else
      jsgr.drawString(mdsValueString(value),35,mdsGraphHeight-average+2);
}

// plot a history graph from stored sparse data
// --------------------------------------------

function mdsHistPlot (which)
{
   if (!mdsUpdateDisplay) return;

   if (!mdsData[which].hist1) return;

   if ((!mdsNodeHistVisible && which.substr(0,5) == 'node_') ||
       (!mdsPerProcHistVisible && which.substr(0,5) == 'proc_'))
   {
      // obviously do not plot if not visible!
      if (mdsData[which].visible)
      {
         mdsData[which].visible = false;
         mdsData[which].hist_jsgr.recreate();
         mdsData[which].hist_jsgr.setPrintable(mdsWzPrintable);
      }
      return;
   }
   mdsData[which].visible = true;

   mdsHistGraph(which);

   if (!mdsData[which].active)
   {
      mdsHistPaint (which);
      return;
   }

   var jsgr = mdsData[which].hist_jsgr;

   // wz canvas begins in the top-left, pen beneath and right of coords
   var scale = mdsData[which].scale;

   mdsHistAverage(jsgr,mdsData[which].ave,scale);

   var scale = mdsData[which].scale;

   var hist1 = mdsData[which].hist1,
       hist2 = mdsData[which].hist2;

   var x = y = lastx1 = lasty1 = lastx2 = lasty2 = 0;
   for (tdata in hist1)
   {
      x = mdsTimeStamp - parseInt(tdata);

      // eliminate stale data
      if (x > mdsHistoryWidth)
      {
         if (x >= mdsHistoryWidth + 10)
         {
            delete hist1[tdata];
            delete hist2[tdata];
            continue;
         }
      }
      if (x > mdsGraphWidth+10) continue;
      if (x > mdsGraphWidth) x = mdsGraphWidth + 1;

      y = hist1[tdata] * mdsGraphHeight / scale;
      if (y > mdsGraphHeight) y = mdsGraphHeight;
      jsgr.setColor(mdsHistCol1);
      if (lasty1 > 0)
         jsgr.drawLine (mdsGraphWidth-lastx1, mdsGraphHeight-lasty1,
                        mdsGraphWidth-x, mdsGraphHeight-lasty1);
      jsgr.drawLine (mdsGraphWidth-x, mdsGraphHeight-lasty1,
                     mdsGraphWidth-x, mdsGraphHeight-y);
      lastx1 = x;
      lasty1 = y;

      if (hist2[tdata] != undefined)
      {
         y = hist2[tdata] * mdsGraphHeight / scale;
         if (y > mdsGraphHeight) y = mdsGraphHeight;
         jsgr.setColor(mdsHistCol2);
         if (lasty2 > 0)
            jsgr.drawLine (mdsGraphWidth-lastx2, mdsGraphHeight-lasty2,
                           mdsGraphWidth-x, mdsGraphHeight-lasty2);
         jsgr.drawLine (mdsGraphWidth-x, mdsGraphHeight-lasty2,
                        mdsGraphWidth-x, mdsGraphHeight-y);
         lastx2 = x;
         lasty2 = y;
      }
   }

   if (lasty1 > 0)
   {
      jsgr.setColor(mdsHistCol1);
      jsgr.drawLine (mdsGraphWidth-lastx1, mdsGraphHeight-lasty1,
                     mdsGraphWidth, mdsGraphHeight-lasty1);
   }
   if (lasty2 > 0)
   {
      jsgr.setColor(mdsHistCol2);
      jsgr.drawLine (mdsGraphWidth-lastx2-1, mdsGraphHeight-lasty2,
                     mdsGraphWidth-1, mdsGraphHeight-lasty2);
   }

   mdsHistScale(jsgr,which,mdsData[which].val1,
                           mdsData[which].val2,
                           mdsData[which].scale);

   mdsHistPaint (which);
}

///////////////////////////
// iframe data streaming //
///////////////////////////

// cancel any outstanding alert and optionally set a new one
// ---------------------------------------------------------

var mdsAlertTimer = null;

function mdsResetAlert (newAlert)
{
   // each minute the alert gets recheduled if updating is occuring
   clearTimeout(mdsAlertTimer);
   mdsAlertTimer = null;
   if (newAlert) mdsAlertTimer = setTimeout("mdsAlert()",mdsAlertSecs*1000);
}

// that the node has stopped being updated
// ---------------------------------------

function mdsAlert ()
{
   // the above timer has expired
   mdsAlertTimer = null;
   var nodeNotes = document.getElementById('node_notes');
   var notesText = nodeNotes.innerHTML;
   nodeNotes.innerHTML = null;
   nodeNotes.innerHTML = mdsAlertText + notesText;
}

// send a message from a child iframe to the parent
// ------------------------------------------------

function mdsToMsgChannel (forEval)
{
   // MSIE(6) and SeaMonkey(1.6) (at least) deny permission for these
   try
   {
      var win = parent.window;
      if (win != undefined && typeof win.postMessage == 'function')
         // message posting browsers (HTML5)
         win.postMessage (forEval, '*');
      else
         // inter-frame kludge
         mdsToBackChannel (forEval);
   }
   catch (err)
   {
      // inter-frame kludge
      mdsToBackChannel (forEval);
   }
}

// message passing via hidden iframe (HTML4)
// -----------------------------------------

var mdsBackChanCount = 0;

function mdsToBackChannel (forEval)
{
   var url = mdsRefererHost + '/mondesi/-/js/bchan.html#' + forEval;
   // MSIE(6+7) gets confused removing then adding elements with identical IDs!
   var iframe = document.getElementById('bchan_iframe_'+mdsBackChanCount);
   if (iframe != undefined)
   {
      mdsBody.removeChild(iframe);
      delete iframe;
   }
   iframe = document.createElement('iframe');
   iframe.setAttribute('id','bchan_iframe_'+(mdsBackChanCount++));
   iframe.setAttribute('src',url);
   iframe.setAttribute('height',0);
   iframe.setAttribute('width',0);
   iframe.setAttribute('frameBorder',0);
   mdsBody.appendChild(iframe);
}

// stop/start the streaming data flow
// ----------------------------------

var mdsDataIframeCount = 0;
var mdsDataIframe = undefined;

function mdsStreamData (start)
{
   if (WASDwebSocket)
   {
      // close any existing connections
      webSocketClose();
      mdsResetAlert (false);
      if (!start) return;

      if (webSocketOpen())
      {
         mdsResetAlert (true);
         return;
      }
      // it's entirely possible the server but not browser supports WebSocket
      WASDwebSocket = false;
   }

   // via iframe and HTML push

   if (mdsDataIframe != undefined)
   {
      // MSIE (6+7) seems to need this to dissociate from the data stream
      mdsDataIframe.src = '/mondesi/-/js/nl.html';

      /* dispose of the iframe dropping the data stream */
      mdsBody.removeChild(mdsDataIframe);
      delete mdsDataIframe;
      mdsDataIframe = undefined;
      mdsResetAlert (false);
   }

   if (start)
   {
      var url = mdsScriptName + '?' +
                (mdsTopProcVisible ? '1' : '0') + '+' + mdsPerProcPid;
      mdsDataIframe = document.createElement('iframe');
      // MSIE(6+7) gets confused removing then adding elements with same IDs!
      mdsDataIframe.setAttribute('id','node_iframe_'+mdsDataIframeCount);
      mdsDataIframe.setAttribute('src',url);
      mdsDataIframe.setAttribute('height',0);
      mdsDataIframe.setAttribute('width',0);
      mdsDataIframe.setAttribute('frameBorder',0);
      mdsBody.appendChild(mdsDataIframe);
      mdsResetAlert (true);
   }
}

//////////////////////////////////////////
// WebSocket support (for WASD v10.1ff) //
//////////////////////////////////////////

var ws = null;
var nodeControl = null;

function webSocketOpen () {

   if (!('WebSocket' in window)) return false;

   if (ws) return true;

   if (window.location.protocol == 'http:')
      var URL = 'ws://';
   else
      var URL = 'wss://';
   URL += window.location.host + mdsScriptName + '?' +
          (mdsTopProcVisible ? '1' : '0') + '+' + mdsPerProcPid;

   ws = new WebSocket(URL);

   ws.onopen = function(evt) {
   };

   ws.onclose = function(evt) {
      if (ws) ws.close();
      ws = null;
      // try to reconnect in a staggered fashion (in case multiple CGIplus)
      var tout = 1500 + Math.floor((Math.random()*1000));
      if (mdsRunning) setTimeout ('webSocketOpen()', tout);
  };

   ws.onmessage = function (evt) { 
      var data = evt.data;
      eval (data);
   };

   return true;
}

function webSocketClose () {

   if (!WASDwebSocket) return;
   if (!('WebSocket' in window)) return;

   if (ws) ws.close();
   ws = null;
}

/*****************************************************************************/
